/*
 * OCSP responder
 * by Massimiliano Pala (madwolf@openca.org)
 * OpenCA project 2001
 *
 * Copyright (c) 2001 The OpenCA Project.  All rights reserved.
 *
 * ====================================================================
 *
 * This product includes cryptographic software written by Eric Young
 * (eay@cryptsoft.com).  This product includes software written by Tim
 * Hudson (tjh@cryptsoft.com).
 *
 */

/* NOTICE:
 *
 * All the db stuff now are obsolete as the version now uses the
 * crl to check the certificate(s). The TXT_DB file will be
 * used in future for the suspended certificate(s) list (not
 * yet revoked but not to be considered valid
 *
 */

#include "general.h"

#ifdef HAVE_ENGINE
#include <openssl/engine.h>
#endif

#include <pwd.h>
#include <grp.h>

#include <time.h>
#include <sys/wait.h>

#include "configuration.h"
#include "support.h"
#include "db.h"
#include "http_request.h"
#include "server.h"
#include "ocspd.h"

/* General exported variables */
int debug = 0;
int verbose = 0;

static char *ocspd_usage[] = {
"OCSPd - OpenCA OCSP responder daemon\n",
"(c) 2002 by OpenCA Group\n",
"\n",
"   USAGE: ocspd args\n",
"\n",
" -d              - Daemon, detach from current console\n",
" -p n            - Start listening on port n [2560]\n",
" -b addr         - Binds to ip <addr> [*]\n",
" -c file         - A config file\n",
" -md digest      - Set digest to be used [md5]\n",
" -k pwd          - Password protecting the private key (if any)\n",
" -i passin       - Passin arg\n",
#ifdef HAVE_ENGINE
" -e engine       - use engine e, possibly a hardware device.\n",
#endif
" -v              - Talk alot while doing things\n",
NULL
};

/* Staic variables */
char *prgname = "ocspd";
OCSPD_CONFIG *ocspd_conf = NULL;
char *shm = NULL;

/* Local functions prototypes */
int writePid ( int pid, char *pidfile );

/* Main */
int main ( int argc, char *argv[] ) {

	struct passwd *pw = NULL;
	struct group *gr = NULL;

	char *usr = NULL;
	char *group = NULL;

	char *port_s  = NULL;
	char *bind_s  = NULL;

	char *child_s = NULL;
	char *queue_s = NULL;
	char *maxrs_s = NULL;
	char *crlUrl_s = NULL;
	char *crlEntry_s = NULL;

	int child_num = 5;
	int queue_num = 100;
	int maxrs_num = 8192;

	int daemon = 0;
	int badops = 0;
	int ret = 0;
	BIO *keyf = NULL;
	BIO *certf = NULL;

	char *keyfile = NULL;
	char *configfile = NULL;
	char *dbfile = NULL;
	char *cert = NULL;
	char *cacert = NULL;
	char **pp = NULL;
	char *key = NULL;
	char *passargin = NULL;
	char *section = NULL;
	char *resp_section = NULL;
	char *dbms_section = NULL;
	char *pidfile = NULL;
	char *tmp_s = NULL;
	int keyform = FORMAT_PEM;

        /* ADD ENGINE SUPPORT */
#ifdef HAVE_ENGINE
        char *engine=NULL;
        ENGINE *e = NULL;
#endif

	int *child_list = NULL;
	int i = 0;
	pid_t pid = 0;
	pid_t ppid = 0;

	char *ocsp_digest_name = NULL;

	int flags = 0;

	/*
	CONF *conf = NULL;
	EVP_PKEY *pkey = NULL;
	TXT_DB *db=NULL;
	*/

	/* What are really needed here ?? Please help... */
	X509V3_add_standard_extensions();
	OpenSSL_add_all_algorithms();

	OpenSSL_add_all_digests();
	OpenSSL_add_all_ciphers();

	argv++;
	argc--;

	openlog( prgname, LOG_PID, LOG_DAEMON);
	while (argc >= 1)
		{
		/* Debuggin options */
#ifdef DEBUG
		syslog(LOG_ERR,"Parsing (%s)\n", *argv);
#endif
                if (strcmp(*argv,"-c") == 0)
                        {
                        if (--argc < 1) goto bad;
                        configfile= *(++argv);
                        }
		else if (strcmp(*argv,"-k") == 0)
			{
			if (--argc < 1) goto bad;
			key= *(++argv);
			}
		else if (strcmp(*argv,"-i") == 0)
			{
			if (--argc < 1) goto bad;
			passargin= *(++argv);
			}
		else if (strcmp(*argv,"-md") == 0)
			{
			if (--argc < 1) goto bad;
			ocsp_digest_name = *(++argv);
			}
#ifdef HAVE_ENGINE
		else if (strcmp(*argv,"-e") == 0)
			{
			if (--argc < 1) goto bad;
			engine = *(++argv);
			}
#endif
		else if (strcmp(*argv,"-p") == 0)
			{
			if (--argc < 1) goto bad;
			port_s = *(++argv);
			}
		else if (strcmp(*argv,"-b") == 0)
			{
			if (--argc < 1) goto bad;
			bind_s = *(++argv);
			}
		else if (strcmp(*argv,"-v") == 0)
			verbose=1;
		else if (strcmp(*argv,"-d") == 0)
			daemon=1;
		else badops = 1;
		argc--;
		argv++;
		}

bad:
	if (badops) {
		
		for (pp=ocspd_usage; (*pp != NULL); pp++)
			printf(*pp);
		goto err;
	}

	ERR_load_crypto_strings();
	if(( ocspd_conf = (OCSPD_CONFIG *) 
			OPENSSL_malloc ( sizeof(OCSPD_CONFIG))) == NULL ) {
		syslog(LOG_ERR, "Memory Allocation error");
		exit(-1);
	}

	/* load configuration file */
	if( configfile == NULL ) 
		configfile = OCSPD_DEF_CONFIG;
	if( (ocspd_conf->conf = load_config( configfile, &section )) == NULL ) {
		syslog( LOG_ERR,"can not load configuration file %s",
				configfile );
		exit(-1);
	}

	/* Copy the conf filename so that it could be re-loaded on SIGHUP */
	if( (ocspd_conf->cnf_filename = 
		(char *) OPENSSL_malloc(strlen(configfile) + 1)) == NULL ) {
		syslog( LOG_ERR,"memory allocation error (%d)", __LINE__ );
		goto err;
	}

	/* Get db Filename */
        if ((dbfile == NULL) && 
		((dbfile = NCONF_get_string(ocspd_conf->conf, section,
					ENV_DATABASE)) == NULL)) {
		goto err;
       	}

	/* load/init certificates db */
	 if(( ocspd_conf->db = ocspd_load_db_file( dbfile )) == NULL ) {
		syslog( LOG_ERR,"can not load/init db (%s)", dbfile );
		goto err;
	}

	/* Copy the db filename so that it could be re-loaded on SIGHUP */
	strcpy( ocspd_conf->cnf_filename, configfile);
	if( (ocspd_conf->db_file = 
		(char *) OPENSSL_malloc(strlen(dbfile) + 1)) == NULL ) {
		syslog( LOG_ERR,"memory allocation error (%d)", __LINE__ );
		goto err;
	}
	strcpy( ocspd_conf->db_file, dbfile);

        /* ENGINE support added */
#ifdef HAVE_ENGINE
	if (engine != NULL) {
		if((e = ENGINE_by_id(engine)) == NULL) {
			syslog(LOG_ERR,"invalid engine \"%s\"", engine);
			goto err;
		}

		if(!ENGINE_set_default(e, ENGINE_METHOD_ALL)) {
			syslog(LOG_ERR,"can't use that engine");
			goto err;
		}
		syslog(LOG_ERR,"engine \"%s\" set.\n", engine);
		ENGINE_free(e);
	}
#endif

        /* Passin support */
        if(!key && !app_passwd(passargin, NULL, &key, NULL)) {
                syslog(LOG_ERR, "Error getting passwords\n");
                goto err;
        }

	/*****************************************************************/
        /* Reading Private key file */
        if ((keyfile == NULL) && 
			((keyfile = NCONF_get_string(ocspd_conf->conf, section,
					ENV_PRIVATE_KEY)) == NULL)) {
                lookup_fail(section,ENV_PRIVATE_KEY);
		goto err;
       	}

        if( verbose )
                syslog(LOG_INFO,"Reading Private Key file %s", keyfile);
        if ((keyf=BIO_new_file( keyfile, "r")) == NULL) {
		syslog( LOG_ERR,"cannot open BIO file, why ?" );
                goto err;
	}

        if ( keyform == FORMAT_ENGINE)
	        {
#ifdef HAVE_ENGINE
		/*
                if(!e) {
			BIO_printf( bio_err,"no engine loaded!\n");
                        goto err;
                }
                pkey = ENGINE_load_private_key( e, keyfile, key );
		*/
                syslog(LOG_ERR,"No engine support yet.\n");
                goto err;
#else
                syslog(LOG_ERR,"No engine support compiled.\n");
                goto err;
#endif
	        } else if ( keyform == FORMAT_PEM ) {
	                ocspd_conf->ocspd_pkey = 
				PEM_read_bio_PrivateKey(keyf,NULL,NULL,key);
	                if (key) memset(key,0,strlen(key));
	        } else {
                syslog(LOG_ERR,"bad input format specified for key file\n");
		goto err;
        };
        if( keyf ) BIO_free( keyf );

        if( ocspd_conf->ocspd_pkey == NULL ) {
                syslog(LOG_ERR,"Error loading private key\n");
                goto err;
        };


	/*****************************************************************/
	/* Load the OCSP certificate file */
        if ((cert == NULL) && ((cert=NCONF_get_string(ocspd_conf->conf,
                section,ENV_CERTIFICATE)) == NULL)) {
                lookup_fail(section,ENV_CERTIFICATE);
        }

	if ( cert ) {
		if( verbose )
	                syslog(LOG_INFO,"reading certificate file.\n");
	        if ((certf=BIO_new_file( cert, "r")) == NULL) {
	                syslog(LOG_ERR,"unable to open certificate file.\n");
	                goto err;
		}
	        if ((ocspd_conf->ocspd_cert = PEM_read_bio_X509(certf,NULL,NULL,NULL))
				==NULL) {
	                syslog(LOG_ERR,"cannot load certificate.\n");
	                goto err;
		}
	        BIO_free( certf );
	}

        if (ocspd_conf->ocspd_cert == NULL) {
               	syslog(LOG_ERR,"unable to load OCSP certificate\n");
               	goto err;
	}

        if (!X509_check_private_key(ocspd_conf->ocspd_cert,ocspd_conf->ocspd_pkey)) {
               	syslog(LOG_ERR, "OCSP cert and private key do not match\n");
               	goto err;
	}

	/*****************************************************************/
	/* Load the CA certificate file */
        if ((cacert == NULL) && ((cacert=NCONF_get_string(ocspd_conf->conf,
                section,ENV_CA_CERTIFICATE)) == NULL))
                {
                lookup_fail(section,ENV_CA_CERTIFICATE);
                goto err;
                }
	if ( cacert ) {
		if( verbose )
	                syslog(LOG_INFO,"reading CA certificate file.\n");
	        if ((certf=BIO_new_file( cacert, "r")) == NULL) {
	                syslog(LOG_ERR,"unable to open CA certificate file.\n");
	                goto err;
		}
	        if ((ocspd_conf->cacert = PEM_read_bio_X509(certf,NULL,NULL,NULL))
				==NULL) {
	                syslog(LOG_ERR,"cannot load CA certificate.\n");
	                goto err;
		}
	        BIO_free( certf );
	}
        if (ocspd_conf->cacert == NULL) {
                syslog(LOG_ERR,"unable to load CA certificate\n");
                goto err;
        }

	if(verbose)
		syslog(LOG_INFO,"OCSP Daemon setup completed");

        /*****************************************************************/
        /* Let's get the digest */
        if ((ocsp_digest_name == NULL) &&
	                ((ocsp_digest_name=NCONF_get_string(ocspd_conf->conf,section,
                                                   ENV_OCSPD_MD)) == NULL)) {
                lookup_fail(section,ENV_OCSPD_MD);
                goto err;
         }

        ocspd_conf->digest = (EVP_MD *) EVP_get_digestbyname(ocsp_digest_name);
        if( ocspd_conf->digest == NULL ) {
                syslog(LOG_ERR, "unsupported digest type %s",
                                                        ocsp_digest_name);
                goto err;
        }

	/*****************************************************************/
	/* Load other important config keys */

	/* Listen to port_s */
        if ((port_s == NULL) && ((port_s = NCONF_get_string(ocspd_conf->conf, section,
					ENV_OCSPD_PORT)) == NULL)) {
               	if( verbose )
			lookup_fail(section,ENV_OCSPD_PORT);
		port_s = "2560";
       	}

	/* Binds to address */
        if ((bind_s == NULL) && ((bind_s = NCONF_get_string(ocspd_conf->conf, section,
					ENV_OCSPD_BIND)) == NULL)) {
               	if( verbose )
			lookup_fail(section,ENV_OCSPD_BIND);
		bind_s = "*";
       	}

	/* Spawn child_num servers */
        if ((child_s == NULL) && ((child_s = NCONF_get_string(ocspd_conf->conf, section,
					ENV_OCSPD_CHILD)) == NULL)) {
               	if( verbose )
			lookup_fail(section,ENV_OCSPD_CHILD);
       	} else {
		child_num = atoi( child_s );
	}

	/* Max clients per server (queue len) */
        if ((queue_s == NULL) && ((queue_s = NCONF_get_string(ocspd_conf->conf, section,
					ENV_OCSPD_QUEUE)) == NULL)) {
               	if( verbose )
			lookup_fail(section,ENV_OCSPD_QUEUE);
       	} else {
		queue_num = atoi( queue_s );
	}

	/* Max req size allowed (http header included) */
        if ((maxrs_s == NULL) && ((maxrs_s = NCONF_get_string(ocspd_conf->conf, section,
					ENV_OCSPD_MAXRS)) == NULL)) {
               	if( verbose )
			lookup_fail(section,ENV_OCSPD_MAXRS);
       	} else {
		maxrs_num = atoi( maxrs_s );
	}

	/* Get the user name to set the process to */
        if ((group == NULL) && ((group = NCONF_get_string(ocspd_conf->conf, section,
					ENV_OCSPD_GROUP)) == NULL)) {
               	if( verbose )
			lookup_fail(section,ENV_OCSPD_GROUP);

		group = OCSPD_DEF_GROUP;
       	}

	/* Get the user name to set the process to */
        if ((usr == NULL) && ((usr = NCONF_get_string(ocspd_conf->conf, section,
					ENV_OCSPD_USER)) == NULL)) {
               	if( verbose )
			lookup_fail(section,ENV_OCSPD_USER);

		usr = OCSPD_DEF_USER;
       	}

	/* Get the pidfile name to write main pid to */
        if ((pidfile == NULL) && ((pidfile = NCONF_get_string(ocspd_conf->conf, section,
					ENV_OCSPD_PIDFILE)) == NULL)) {
               	if( verbose )
			lookup_fail(section,ENV_OCSPD_PIDFILE);

		usr = OCSPD_DEF_PIDFILE;
       	}
	/*****************************************************************/
	/* OCSPD crl access */

	/* Get CRL section name */
        if ((dbms_section = NCONF_get_string(ocspd_conf->conf, section,
					ENV_OCSPD_DBMS)) == NULL) {
               	if( verbose )
			lookup_fail(section, ENV_OCSPD_DBMS);
		dbms_section = ENV_DEFAULT_OCSPD;
       	}

	/* Get the crl URL */
        if ((crlUrl_s = NCONF_get_string(ocspd_conf->conf, dbms_section,
					ENV_OCSPD_CRL_URL)) == NULL) {
               	if( verbose )
			lookup_fail(dbms_section, ENV_OCSPD_CRL_URL);
       	} else {
		ocspd_conf->crl_data.url = getParsedUrl(crlUrl_s);
		if( ! ocspd_conf->crl_data.url ) {
			syslog(LOG_ERR, "ERROR in CRL url specification %s",
					crlUrl_s);
			goto err;
		}
	}

	/* Get the CRL entry, usually the Base DN of the LDAP */
	if( ocspd_conf->crl_data.url->proto == OCSP_CRL_PROTO_LDAP ) {
		if( verbose )
			syslog(LOG_ERR,"Using LDAP protocol for CRL retrivial");
        	if ((crlEntry_s = NCONF_get_string(ocspd_conf->conf, dbms_section,
					ENV_OCSPD_CRL_ENTRY_DN)) == NULL) {
        	       	if( verbose )
				lookup_fail(dbms_section, ENV_OCSPD_CRL_URL);
       		} else {
			ocspd_conf->crl_data.url->dn = (crlEntry_s);
		}
	}

	/* load CRL data */
	ocspd_conf->crl_data.crl = ocspd_get_crl(ocspd_conf->crl_data.url);

	if( !ocspd_conf->crl_data.crl ) {
		syslog( LOG_ERR, "Cannot access CRL data (%s)", crlUrl_s );
			goto err;
	}

	/*****************************************************************/
	/* OCSPD responder specific options */
	ocspd_conf->flags = 0;

	/* Get response section name */
        if ((resp_section = NCONF_get_string(ocspd_conf->conf, section,
					ENV_OCSPD_RESPONSE)) == NULL) {
               	if( verbose )
			lookup_fail(section, ENV_OCSPD_RESPONSE);
		resp_section = ENV_DEFAULT_OCSPD;
       	}

	/* Get if the response has to include keyid */
        if ((tmp_s = NCONF_get_string(ocspd_conf->conf, resp_section,
					ENV_OCSPD_RESPONSE_KEYID)) == NULL) {
               	if( verbose )
			lookup_fail(resp_section, ENV_OCSPD_RESPONSE_KEYID);
       	} else {
		if( strstr( tmp_s, "n" ) || strstr( tmp_s, "N" )) {
			ocspd_conf->flags |= OCSP_RESPID_KEY;
		}
	}

	/* Include in the response the other_certs X509 STACK */
        if ((tmp_s = NCONF_get_string(ocspd_conf->conf, resp_section,
				ENV_OCSPD_RESPONSE_OTHER_CERTS)) == NULL) {
               	if( verbose )
			lookup_fail(resp_section, ENV_OCSPD_RESPONSE_OTHER_CERTS);
		ocspd_conf->flags |= OCSP_NOCERTS;
       	} else {
		ocspd_conf->other_certs = load_certs( NULL, tmp_s,
			FORMAT_PEM, NULL, "responder other certificates" );
	}

	/* Get next update days */
        if ((tmp_s = NCONF_get_string(ocspd_conf->conf, resp_section,
					ENV_OCSPD_RESPONSE_DAYS)) == NULL) {
               	if( verbose )
			lookup_fail(resp_section, ENV_OCSPD_RESPONSE_DAYS);
       	} else {
		ocspd_conf->ndays = atoi( tmp_s );
	}

	/* Get next update minutes */
        if ((tmp_s = NCONF_get_string(ocspd_conf->conf, resp_section,
					ENV_OCSPD_RESPONSE_MINS)) == NULL) {
               	if( verbose )
			lookup_fail(resp_section, ENV_OCSPD_RESPONSE_MINS);
       	} else {
		ocspd_conf->nmin = atoi( tmp_s );
	}

	/*****************************************************************/
	/* Let's have the program running under username.groupid, somewhere */

	if( (gr = getgrnam( group ) ) == NULL ) {
		syslog( LOG_ERR, "Cannot find group %s", group);
		goto err;
	}
	
	if (setgid (gr->gr_gid) == -1) {
		syslog(LOG_ERR,"Error setting group %d (%s)", gr->gr_gid, group);
		exit (1);
	}

	if( (pw = getpwnam( usr ) ) == NULL ) {
		syslog( LOG_ERR, "Cannot find user %s", usr);
		goto err;
	}

	if (setuid (pw->pw_uid) == -1) {
		syslog(LOG_ERR,"Error setting user %d (%s)", pw->pw_uid, usr );
		exit (1);
	}

	/*****************************************************************/
	/* Main spawn and signal routines */

	if (verbose)
		syslog(LOG_INFO,"Configuration loaded and parsed");

	if( daemon ) {
		pid = fork();
		if( pid == 0 ) {
			/* Main process, we have to save the pid to the
			 * pidfile and then exit */
			writePid( getpid(), pidfile );
		} else if ( pid > 0 ) {
			/* Nop */
			goto end;
		} else {
			syslog( LOG_ERR, "Error While spawning child %d", i );
			goto err;
		}
	} else {

		ppid = getpid();
		writePid( ppid, pidfile );
	}


	start_server( bind_s, port_s, child_num, 
				queue_num, maxrs_num, ocspd_conf );

	goto end;


err:
	ret = -1;

end:
	/*
	if( X509_cert ) X509_free( X509_cert );
	if( X509_cacert ) X509_free( X509_cacert );
	if( conf ) NCONF_free( conf );
	if( db ) TXT_DB_free( db );
	if( pkey ) EVP_PKEY_free( pkey );
	*/

	return (ret);
}

int writePid ( int pid, char *pidfile ) {
	FILE *fd;

	if( (fd = fopen( pidfile, "w" )) == 0 ) {
		syslog( LOG_ERR, "Cannot open pidfile (%s)",
				pidfile );
		return(0);
	}

	fprintf( fd, "%d", getpid());
	fclose( fd );

	return(1);
}

